#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_uart.h"

#include "driverlib/fpu.h"
#include "driverlib/debug.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/uart.h"
#include "utils/uartstdio.h"

#include "inc/lm4f232h5qd.h"

#include <stdio.h>
#include <string.h>
#include <alloca.h>

#include "amx.h"
#include "terminal.h"

#define UART_BASE			UART0_BASE
#define UART_PERIPH			SYSCTL_PERIPH_UART0
#define UART_INT			INT_UART0

#define UART_PIN_BASE		GPIO_PORTA_BASE
#define UART_PIN_PERIPH		SYSCTL_PERIPH_GPIOA
#define UART_TX_PIN			GPIO_PIN_1
#define UART_RX_PIN			GPIO_PIN_0

#include "hello_dat.h"
//#include "float_test_dat.h"

/* PawnVMpo͊֐ */
int amx_putstr(char* s) {
	int len = strlen(s);
	UARTprintf("%s",s);
	return len;
}

/* PawnVMpo͊֐ */
int amx_putchar(char c) {
	MAP_UARTCharPut(UART_BASE , c);
	return 1;
}

/* PawnVMpR\[tbV֐ */
int amx_fflush(){
  return 0;
}

/* PawnVMp͊֐ */
int amx_getch(){
  return  0;
}

/* native֐Xg */
const AMX_NATIVE_INFO MY_Natives[] = 
{
	{ "printf"             , n_printf },
	{ NULL , NULL }        // terminator
};

/* PawnVM{ */
int amx_proc(){
	int err;
	AMX amx;
	cell ret;
	AMX_HEADER hdr;
	void *prog;

	UARTprintf("Pawn VM Init\r\n");

	memcpy(&hdr, prog_data ,  sizeof(AMX_HEADER));	/* wb_[̎o */
	UARTprintf("\r\nstp=%d size=%d\r\n\r\n",(int)hdr.stp,(int)hdr.size);

	if (hdr.magic != AMX_MAGIC) {
		UARTprintf("ByteCode Error!!!\r\n");	/* oCgR[hs */
		return -1;
	}

	prog = alloca(hdr.stp);					/* VMp̍Ɨ̈̊m */
	memcpy(prog, prog_data ,  hdr.size);	/* vOf[^̃Rs[ */

	memset(&amx, 0, sizeof(amx));
	err = amx_Init( &amx, prog );				/* VM̏ */
	err = amx_Register(&amx, MY_Natives, -1);	/* native֐̓o^ */
	amx_FloatInit(&amx);						/* FloatW[̓o^ */
	
	UARTprintf("Pawn VM Exec\r\n");
	err = amx_Exec(&amx, &ret, AMX_EXEC_MAIN);	/* VM̎s */
	amx_Cleanup(&amx);							/* VM̔j */
	
	if (err != AMX_ERR_NONE){
		UARTprintf("Run time error %d\r\n", err);
	}
	else {
		UARTprintf("Done!!!!\r\n");
	}

	return err;
}

int main(void)
{
	/* LM4FFPU͕_vZsOɗLɂĂ */
	FPUEnable();
	FPULazyStackingEnable();
	
	IntMasterEnable();
	
	/* VXeNbN50MHzɐݒ */
	ROM_SysCtlClockSet(SYSCTL_SYSDIV_4 | SYSCTL_USE_PLL | SYSCTL_OSC_MAIN | SYSCTL_XTAL_16MHZ);

	/* UART̃n[hEFAݒ */
	MAP_SysCtlPeripheralEnable(UART_PERIPH);
	MAP_SysCtlPeripheralEnable(UART_PIN_PERIPH);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_TX_PIN);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_RX_PIN);
	UARTStdioInit(0);

	UARTprintf("\r\nPawnVM Step0 \r\n");

	amx_proc();

	UARTprintf("\r\nPawnVM Stop \r\n");
}
