#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_uart.h"

#include "driverlib/fpu.h"
#include "driverlib/debug.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/uart.h"
#include "utils/uartstdio.h"

#include "inc/lm4f232h5qd.h"

#include <stdio.h>
#include <string.h>
#include <alloca.h>

#include "usb_cdc_driver.h"
#include "amx.h"
#include "terminal.h"

#define UART_BASE			UART0_BASE
#define UART_PERIPH			SYSCTL_PERIPH_UART0
#define UART_INT			INT_UART0

#define UART_PIN_BASE		GPIO_PORTA_BASE
#define UART_PIN_PERIPH		SYSCTL_PERIPH_GPIOA
#define UART_TX_PIN			GPIO_PIN_1
#define UART_RX_PIN			GPIO_PIN_0

#define VMAck       '\x01'
#define VMCSError   '\x02'
#define VMSizeError '\x03'
#define VMProgEnd   '\x04'
#define VMAbort     '\x05'
#define VMRTError   '\x06'

#define VMIdle              0x1
#define VMSizeOK            0x2
#define VMSizeCSError       0x3
#define VMByteCodeOK        0x4
#define VMByteCodeCSError   0x5
#define VMByteCodeError     0x6
#define VMRun               0x8
#define VMInterrupt         0x10

/* fBC֐[msP] */
void mwait (unsigned int pause_ms)
{
	while(pause_ms--)
	{
		MAP_SysCtlDelay( ( (MAP_SysCtlClockGet() / 3) / 1000 ) );
	}
}

/* PawnVMpo͊֐ */
int amx_putstr(char* s) {
	int len = strlen(s);
	USBCDCBufferWrite( (unsigned char *)s, len);
	return len;
}

/* PawnVMpo͊֐ */
int amx_putchar(char c) {
	unsigned char outChar = c;
	USBCDCBufferWrite( &outChar, 1);
	return 1;
}

/* PawnVMpR\[tbV֐ */
int amx_fflush(){
  return 0;
}

/* PawnVMp͊֐ */
int amx_getch(){
  return  0;
}

/* native֐Xg */
const AMX_NATIVE_INFO MY_Natives[] = 
{
	{ "printf"             , n_printf },
	{ NULL , NULL }        // terminator
};

/* PawnVM{ */
int amx_proc(){
	int err;
	AMX amx;
	cell ret;
	AMX_HEADER hdr;
	void *prog;
	unsigned char len_h , len_l , checkSUM;
	unsigned int sum , i;
	int  p_len ;
	unsigned char *prog_data;
	unsigned char mesBuffer[3];
	unsigned long data_count;

	/* 3oCgM܂ő҂ */
	do{
		data_count = USBCDCBufferDataAvailable();
	}while(data_count < 3);
	
	USBCDCBufferRead(&len_l, 1);	/* oCgR[hiʃoCgjM */
	USBCDCBufferRead(&len_h, 1);	/* oCgR[hiʃoCgjM */
	USBCDCBufferRead(&checkSUM, 1);	/* `FbNTM */

	sum = (len_h + len_l) & 0xFF;	/* `FbNŤvZ */
	if( sum == checkSUM) {
		mesBuffer[0] = VMAck;		/* Ack̑M */
	}
	else {
		mesBuffer[0] = VMSizeCSError;	/* `FbNTG[ */
	}
	mesBuffer[1] = len_l;
	mesBuffer[2] = len_h;

	/* oCgR[h̑M */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
	if(mesBuffer[0] != VMAck ) return VMSizeCSError;	/* G[ꍇ͔ */

	p_len = (len_h << 8) | len_l;	/* oCgR[ȟvZ */
	prog_data = alloca(p_len);		/* oCgR[hi[̈̊m */

	/* oCgR[h̎M */
	data_count=0;
	do
	{
		if(USBCDCBufferDataAvailable() > 0)
		{
			USBCDCBufferRead(&prog_data[data_count], 1);
			data_count++;
		}
	}while(data_count < p_len);

	/* `FbNTM */
	do{
		data_count = USBCDCBufferDataAvailable();
	}while(data_count < 1);
	USBCDCBufferRead(&checkSUM, 1);

	sum = 0;
	for(i=0 ; i< p_len ; i++) sum += prog_data[i];	/* `FbNŤvZ */
	if( (sum & 0xFF) == checkSUM) {
		mesBuffer[0] = VMAck;		/* Ack̑M */
	}
	else {
		mesBuffer[0] = VMSizeCSError;	/* `FbNTG[ */
	}
	
	/* 󋵂 */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 1);
	if(mesBuffer[0] != VMAck ) return VMByteCodeCSError;	/* G[ꍇ͔ */

	memcpy(&hdr, prog_data ,  sizeof(AMX_HEADER));	/* wb_[̎o */

	prog = alloca(hdr.stp);	/* VMp̍Ɨ̈̊m */
	memcpy(prog, prog_data ,  hdr.size);	/* vOf[^̃Rs[ */

	if (hdr.magic != AMX_MAGIC) {
		mesBuffer[0] = VMByteCodeError;		/* oCgR[hG[ */
	}
	else mesBuffer[0] = VMAck;		/* Ack̑M */
	
	mesBuffer[1] = ((int)hdr.stp) & 0xFF;		/* Ɨ̈TCYiʁj */
	mesBuffer[2] = ((int)hdr.stp >> 8) & 0xFF ;	/* Ɨ̈TCYiʁj */

	/* X^bNTCY𑗐M */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
	if(mesBuffer[0] != VMAck ) return VMByteCodeError;	/* G[ꍇ͔ */

	memset(&amx, 0, sizeof(amx));
	err = amx_Init( &amx, prog );				/* VM̏ */
	err = amx_Register(&amx, MY_Natives, -1);	/* native֐̓o^ */
	amx_FloatInit(&amx);						/* FloatW[̓o^ */

	mwait(50);

	err = amx_Exec(&amx, &ret, AMX_EXEC_MAIN);	/* VM̎s */
	amx_Cleanup(&amx);							/* VM̔j */
	
	if (err != AMX_ERR_NONE){
		/* runtimeG[̑M */
		mesBuffer[0] = VMRTError;	/* runntime error */
		mesBuffer[1] = err;			/* error nȏM       */
		mesBuffer[2] = VMRTError;	/* runntime error */
		USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
	}
	else {
		mesBuffer[0] = VMProgEnd;		/* vOIƂzXgAvɓ` */
		USBCDCBufferWrite( (unsigned char *)mesBuffer, 1);
	}

	mwait(200);

	return err;
}

int main(void)
{
	tBoolean USBDeviceValid = false;
	
	/* LM4FFPU͕_vZsOɗLɂĂ */
	FPUEnable();
	FPULazyStackingEnable();
	
	/* VXeNbN50MHzɐݒ */
	ROM_SysCtlClockSet(SYSCTL_SYSDIV_4 | SYSCTL_USE_PLL | SYSCTL_OSC_MAIN | SYSCTL_XTAL_16MHZ);

	IntMasterEnable();

	/* UART̃n[hEFAݒ */
	MAP_SysCtlPeripheralEnable(UART_PERIPH);
	MAP_SysCtlPeripheralEnable(UART_PIN_PERIPH);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_TX_PIN);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_RX_PIN);
	UARTStdioInit(0);

	/* CDC driveȑ */
	USBDeviceValid = USBCDCInit();

	if(USBDeviceValid == false){
		UARTprintf("CDC device invalid \r\n");
	}
	else{
		UARTprintf("Init CDC device\r\n");
	}

	UARTprintf("Pawn Test Ver 1.0 for LM4F Step1 \r\n");

	while(1) {
		amx_proc();
	}
}

