#include "inc/hw_ints.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_gpio.h"
#include "inc/hw_sysctl.h"

#include "driverlib/debug.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/timer.h"
#include "driverlib/usb.h"

#include "usblib/usblib.h"
#include "usblib/usbcdc.h"
#include "usblib/usb-ids.h"
#include "usblib/device/usbdevice.h"
#include "usblib/device/usbdcdc.h"

#include "utils/ustdlib.h"
#include "usb_serial_structs.h"

#include <string.h>
#include <stdarg.h>

static volatile tBoolean g_bUSBConfigured = false;
static void *g_pDevice;

static void (*rxExtHandler)(void);

tBoolean USBCDCInit(){
	/* USB̃n[hEFAݒ */
	ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);
	ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOG);
	ROM_GPIOPinConfigure(GPIO_PG4_USB0EPEN);
	ROM_GPIOPinTypeUSBDigital(GPIO_PORTG_BASE, GPIO_PIN_4);
	ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOL);
	ROM_GPIOPinTypeUSBAnalog(GPIO_PORTL_BASE, GPIO_PIN_6 | GPIO_PIN_7);
	ROM_GPIOPinTypeUSBAnalog(GPIO_PORTB_BASE, GPIO_PIN_0 | GPIO_PIN_1);
	
	// rWA1̏ꍇVBUSv_E
	if(CLASS_IS_BLIZZARD && REVISION_IS_A1)
	{
	    HWREG(GPIO_PORTB_BASE + GPIO_O_PDR) |= GPIO_PIN_1;
	}

	USBBufferInit((tUSBBuffer *)&g_sTxBuffer);
	USBBufferInit((tUSBBuffer *)&g_sRxBuffer);

	/* USBfoCX[h֐؂ւ */
	USBStackModeSet(0, USB_MODE_DEVICE, 0);

	/* CDCfoCX̏ */
	g_pDevice = USBDCDCInit(0, (tUSBDCDCDevice *)&g_sCDCDevice);
	rxExtHandler = NULL;

	if(g_pDevice == NULL){
		return false;
	}
	else{
		return true;
	}
}

tBoolean USBCDCIsConfigured(){
	return g_bUSBConfigured;
}

void USBCDCIntRxExtRegister(void (*handler_func)(void)) {
	rxExtHandler = handler_func;
}

void USBCDCIntRxExtUnregister() {
	rxExtHandler = NULL;
}

void USBCDCBufferFlush(){
	USBBufferFlush(&g_sTxBuffer);
}

void USBCDCRXBufferClear(){
	USBBufferFlush(&g_sRxBuffer);
}


unsigned long USBCDCBufferWrite(unsigned char *pucData, unsigned long ulLength){
	unsigned long txSpace;
	do{
		txSpace = USBBufferSpaceAvailable((tUSBBuffer *)&g_sTxBuffer);
	}while(txSpace == 0 );

	return USBBufferWrite(&g_sTxBuffer, pucData, ulLength);
}

unsigned long USBCDCBufferRead(unsigned char *pucData, unsigned long ulLength){
	return USBBufferRead(&g_sRxBuffer, pucData, ulLength);
}

unsigned long USBCDCBufferDataAvailable(){
	return USBBufferDataAvailable(&g_sRxBuffer);
}

unsigned long USBCDCBufferSpaceAvailable(){
	return  USBBufferSpaceAvailable((tUSBBuffer *)&g_sTxBuffer);
}

void USBCDCSendCR(){
	unsigned char mesBuffer[2];
	
	mesBuffer[0] = '\r';
	mesBuffer[1] = '\n';
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 2);
}

void USBCDCPrintf(const char *formatStr, ...)
{
	va_list arg;
	char mesBuffer[80];

	va_start(arg, formatStr);
	uvsnprintf(mesBuffer, sizeof(mesBuffer), formatStr, arg);
	va_end(arg);
	
	USBCDCBufferWrite( (unsigned char *)mesBuffer, strlen(mesBuffer));
}

unsigned long ControlHandler(void *pvCBData, unsigned long ulEvent, unsigned long ulMsgValue, void *pvMsgData)
{
	switch(ulEvent)
	{
		case USB_EVENT_CONNECTED:
			g_bUSBConfigured = true;
			
			USBBufferFlush(&g_sTxBuffer);
			USBBufferFlush(&g_sRxBuffer);
			break;

		case USB_EVENT_DISCONNECTED:
			g_bUSBConfigured = false;
			break;

		case USBD_CDC_EVENT_GET_LINE_CODING:
			((tLineCoding *)pvMsgData)->ulRate = 115200;
			((tLineCoding *)pvMsgData)->ucDatabits = 8;
			((tLineCoding *)pvMsgData)->ucParity = USB_CDC_PARITY_NONE;
			((tLineCoding *)pvMsgData)->ucStop = USB_CDC_STOP_BITS_1;
			break;
		
		case USBD_CDC_EVENT_SET_LINE_CODING:
		case USBD_CDC_EVENT_SET_CONTROL_LINE_STATE:
		case USBD_CDC_EVENT_SEND_BREAK:
		case USBD_CDC_EVENT_CLEAR_BREAK:
		case USB_EVENT_SUSPEND:
		case USB_EVENT_RESUME:
		    break;

		default:
		    break;
	}
	
	return(0);
}

unsigned long TxHandler(void *pvCBData, unsigned long ulEvent, unsigned long ulMsgValue, void *pvMsgData)
{
	switch(ulEvent)
	{
		case USB_EVENT_TX_COMPLETE:
		    break;

		default:
		    break;
	}
	return(0);
}

unsigned long RxHandler(void *pvCBData, unsigned long ulEvent, unsigned long ulMsgValue, void *pvMsgData)
{
	switch(ulEvent)
	{
		case USB_EVENT_RX_AVAILABLE:
		{
			if( rxExtHandler != NULL ) rxExtHandler();
			return(0);
		}

		case USB_EVENT_DATA_REMAINING:
		{
		    return(0);
		}

		case USB_EVENT_REQUEST_BUFFER:
		{
		    return(0);
		}
		
		default:
		    break;
	}
	
	return(0);
}

