#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_uart.h"

#include "driverlib/fpu.h"
#include "driverlib/debug.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/uart.h"
#include "utils/uartstdio.h"

#include "inc/lm4f232h5qd.h"
#include "fatfs/src/ff.h"
#include "fatfs/src/diskio.h"

#include <stdio.h>
#include <string.h>
#include <alloca.h>

#include "amx.h"
#include "terminal.h"

#define UART_BASE			UART0_BASE
#define UART_PERIPH			SYSCTL_PERIPH_UART0
#define UART_INT			INT_UART0

#define UART_PIN_BASE		GPIO_PORTA_BASE
#define UART_PIN_PERIPH		SYSCTL_PERIPH_GPIOA
#define UART_TX_PIN			GPIO_PIN_1
#define UART_RX_PIN			GPIO_PIN_0

/* fBC֐[msP] */
void mwait (unsigned int pause_ms)
{
	while(pause_ms--)
	{
		MAP_SysCtlDelay( ( (MAP_SysCtlClockGet() / 3) / 1000 ) );
	}
}

/* PawnVMpo͊֐ */
int amx_putstr(char* s) {
	int len = strlen(s);
	UARTprintf("%s",s);
	return len;
}

/* PawnVMpo͊֐ */
int amx_putchar(char c) {
	MAP_UARTCharPut(UART_BASE , c);
	return 1;
}

/* PawnVMpR\[tbV֐ */
int amx_fflush(){
  return 0;
}

/* PawnVMp͊֐ */
int amx_getch(){
  return  0;
}

/* w肵[msP]x */
static cell AMX_NATIVE_CALL n_mwait(AMX *amx,const cell *params)
{
	unsigned int pause_ms = (unsigned int)params[1];
	mwait(pause_ms);
	return 0;
}

/* w肵ytFLɂ */
static cell AMX_NATIVE_CALL n_SysCtlPeripheralEnable (AMX *amx,const cell *params)
{
	unsigned long ulPeripheral = (unsigned long)params[1];
	MAP_SysCtlPeripheralEnable( ulPeripheral );
	
	return 0;
}

/* w肵sGPIOo͂ƂĐݒ */
static cell AMX_NATIVE_CALL n_PinTypeGPIOOutput(AMX *amx,const cell *params)
{
	unsigned long ulPort     = (unsigned long)params[1];
	unsigned char ucPins     = (unsigned char)params[2];
	unsigned long ulStrength = (unsigned long)params[3];
		
	MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_OUT);
	MAP_GPIOPadConfigSet( ulPort , ucPins , ulStrength , GPIO_PIN_TYPE_STD );

	return 0;
}

/* w肵sGPIO͂ƂĐݒ */
static cell AMX_NATIVE_CALL n_PinTypeGPIOInput(AMX *amx,const cell *params)
{
	unsigned long ulPort    = (unsigned long)params[1];
	unsigned char ucPins    = (unsigned char)params[2];
	unsigned long ulPinType = (unsigned long)params[3];

	MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_IN);
	MAP_GPIOPadConfigSet( ulPort , ucPins , GPIO_STRENGTH_2MA, ulPinType );

	return 0;
}

/* w肵s̏Ԃݒ */
static cell AMX_NATIVE_CALL n_GPIOWrite(AMX *amx,const cell *params)
{
	unsigned long ulPort = (unsigned long)params[1];
	unsigned char ucPins = (unsigned char)params[2];
	unsigned char ucVal  = (unsigned char)params[3];

	MAP_GPIOPinWrite( ulPort , ucPins , ucVal );

	return 0;
}

/* w肵s̏Ԃǂݏo */
static cell AMX_NATIVE_CALL n_GPIORead(AMX *amx,const cell *params)
{
	unsigned long ulPort    = (unsigned long)params[1];
	unsigned char ucPins    = (unsigned char)params[2];

	return MAP_GPIOPinRead( ulPort , ucPins );
}

/* native֐Xg */
const AMX_NATIVE_INFO MY_Natives[] = 
{
	{ "printf"                  , n_printf },
	{ "mwait"                   , n_mwait },
	{ "SysCtlPeripheralEnable"  , n_SysCtlPeripheralEnable },
	{ "PinTypeGPIOOutput"       , n_PinTypeGPIOOutput },
	{ "PinTypeGPIOInput"        , n_PinTypeGPIOInput },
	{ "GPIOWrite"               , n_GPIOWrite },
	{ "GPIORead"                , n_GPIORead },
	{ NULL , NULL }        // terminator
};

/* PawnVM{ */
int amx_proc(){
	int err;
	AMX amx;
	cell ret;
	AMX_HEADER hdr;
	void *prog;
	unsigned char *prog_data;	/* oCgR[hi[pz */

	FATFS fatfs;				/* FATt@CVXe */
	FIL   fp;					/* t@CIuWFNg */
	unsigned short bytesRead;

	/* microSDJ[h}Eg */
	if( f_mount(0 , &fatfs) != FR_OK ){
		UARTprintf("mount error\r\n");
		return 0;
	}
	
	/* oCgR[ht@CI[v */
	if( f_open( &fp , "boot.amx" , FA_READ | FA_OPEN_EXISTING ) != FR_OK ){
		UARTprintf("boot.amx not found\r\n");
		return 0;
	}

	UARTprintf("Pawn VM Init\r\n");

	/* oCgR[h̃wb_ǂݏo */
	f_read( &fp , &hdr , sizeof(AMX_HEADER) , &bytesRead );
	UARTprintf("\r\nstp=%d size=%d\r\n\r\n",(int)hdr.stp,(int)hdr.size);

	if (hdr.magic != AMX_MAGIC) {
		UARTprintf("ByteCode Error!!!\r\n");	/* oCgR[hs */
		return -1;
	}

	UARTprintf("code size = %d\r\n",hdr.size);
	prog_data = alloca(hdr.size);				/* oCgR[hi[pz̊m */
	
	f_lseek( &fp , 0 );	/* ǂݏoʒu擪֖߂ */
	f_read( &fp , prog_data , hdr.size , &bytesRead ); /* oCgR[hǂݏo */
	f_close(&fp); /* t@CN[Y */

	prog = alloca(hdr.stp);					/* VMp̃X^bN̈̊m */
	memcpy(prog, prog_data ,  hdr.size);	/* vOf[^̃Rs[ */

	memset(&amx, 0, sizeof(amx));
	err = amx_Init( &amx, prog );				/* VM̏ */
	err = amx_Register(&amx, MY_Natives, -1);	/* native֐̓o^ */
	amx_FloatInit(&amx);						/* FloatW[̓o^ */
	
	UARTprintf("Pawn VM Exec\r\n");
	err = amx_Exec(&amx, &ret, AMX_EXEC_MAIN);	/* VM̎s */
	amx_Cleanup(&amx);							/* VM̔j */
	
	if (err != AMX_ERR_NONE){
		UARTprintf("Run time error %d\r\n", err);
	}
	else {
		UARTprintf("Done!!!!\r\n");
	}

	return err;
}

#include <math.h>

int main(void)
{
	/* LM4FFPU͕_vZsOɗLɂĂ */
	FPUEnable();
	FPULazyStackingEnable();
	
	IntMasterEnable();
	
	/* VXeNbN50MHzɐݒ */
	ROM_SysCtlClockSet(SYSCTL_SYSDIV_4 | SYSCTL_USE_PLL | SYSCTL_OSC_MAIN | SYSCTL_XTAL_16MHZ);

	/* UART̃n[hEFAݒ */
	MAP_SysCtlPeripheralEnable(UART_PERIPH);
	MAP_SysCtlPeripheralEnable(UART_PIN_PERIPH);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_TX_PIN);
	MAP_GPIOPinTypeUART(UART_PIN_BASE, UART_RX_PIN);
	UARTStdioInit(0);

	/* SDJ[h̏ */
	MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_SSI0);

	UARTprintf("\r\nPawnVM Step4 SD_CARD BOOT \r\n");

	amx_proc();

	UARTprintf("\r\nPawnVM Stop \r\n");
}

